import types
import re
import time, datetime
import sys, os

# insert path for stats_cfg
# this variable will be replaced by setup install.
CONFIG_PATH = '/etc/fred'
PURELIB_PATH = '/usr/share/python-support'

if CONFIG_PATH and CONFIG_PATH not in sys.path:
    sys.path.append(CONFIG_PATH)

if PURELIB_PATH and PURELIB_PATH not in sys.path:
    sys.path.append(PURELIB_PATH)

import stats_cfg as config
from dbconn import DBConn

log_file = open(config.LOG_FILE, 'a')


def time_measure(func, *args, **kwd):
    ''' Decorator which measures time of its inner function
        and write result to log file and if DEBUG is True that to stdout, too
    '''
    t1 = time.time()
    result = func(*args, **kwd)
    t_diff = time.time() - t1

    if isinstance(func, types.FunctionType):
        f_name = func.func_name
    else:
        f_name = func.im_func.func_name

    log_file.write('%-40s %10.3f\n' % (f_name + ':', t_diff))
    if config.DEBUG:
        print '%-40s %10.3f' % (f_name + ':', t_diff)

    return result


class AddData(object):
    def __init__(self):
        self.db = DBConn(0)
        self.db_fstats = DBConn(1)
        self.dblog = DBConn(2)
        self.date = None
        self.midnight = None # midnight at the end of day self.date (so we dont have to cast type every comparsion to ::date in SQL)
        self.def_pdict = {}
        self.year, self.month, self.day = None, None, None

    def data_to_insert(self, func):
        ''' Decorator which creates sql INSERT to stat table from data returned by inner function'''

        #create insert commands from data:
        output = ''
        data = time_measure(func)
        for row in data:
            #print row
            #vals = ', '.join([if isinstance(col, strin)str(col) for col in row])
            output += '''INSERT INTO stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) VALUES (%s, %s, %s, '%s', %s, %s, %s, '%s');\n''' % tuple(row)
        return output

    def get_data(self, date_str, compute_sums_this_year=True, compute_all_sums=False, call_functions=None,
                 check_registrars=True, result_in_trasaction=True):
        if check_registrars:
            self.check_and_update_for_new_registrars()
            self.check_registrars()

        self.def_pdict = {}
        self.def_pdict['year'], self.def_pdict['month'], self.def_pdict['day'] = self.year, self.month, self.day = date_str.split('-')
        self.def_pdict['date'] = date_str
        tommorow = datetime.date(*[int(item) for item in date_str.split('-')]) + datetime.timedelta(1)
        self.def_pdict['midnight'] = "'%d-%02d-%02d'::timestamp" % (tommorow.year, tommorow.month, tommorow.day) # midnight at the end of day of 'date'
        self.def_pdict['monitoring_domain'] = config.MONITORING_DOMAIN
        self.def_pdict['monitoring_cltrid'] = config.MONITORING_CLTRID
        self.def_pdict['log_req_cond'] = """
            r.time_begin BETWEEN '%(date)s'::timestamp AND %(midnight)s
            AND NOT r.is_monitoring
            AND r.service_id = 3 -- epp
        """ % self.def_pdict

        if config.TURN_OFF_BITMAP_SCAN:
            self.db.execute('set enable_bitmapscan to off;')

        output = ''
        ##output += 'COPY stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) FROM stdin;\n'
        if call_functions is None:
            call_functions = [
                self.get_number_of_domains,
                self.get_number_of_domains_in_zone,
                self.get_number_of_domains_with_dnssec,
                self.get_domains_in_protection,

                self.get_number_of_other_objects,
                self.get_number_of_deleted_other_objects,

                self.get_number_of_registrars_domains,
                self.get_number_of_registrars_secure_domains,

                self.get_number_of_registrations_by_zone,
                self.get_number_of_deletions,
                self.get_domains_entered_protection,

                self.get_operations_in_registry,
                self.get_operations_errors,
            ]

        for func in call_functions:
            output += self.data_to_insert(func)

        if compute_all_sums:
            output += open(os.path.join(PURELIB_PATH, 'fred_stats', 'compute_all_sums.sql')).read()
        elif compute_sums_this_year:
            output += open(os.path.join(PURELIB_PATH, 'fred_stats', 'compute_sums_this_year.sql')).read()

        if result_in_trasaction and output:
            output = 'BEGIN;\n\n%s\n\nCOMMIT;\n' % output
        return output

    def check_and_update_for_new_registrars(self):
        query = """SELECT MAX(id) FROM registrar;"""
        db_reg_max_id = self.db.fetchquery(query)[0][0]
        query = """SELECT MAX(object_id) FROM enum_rec_type WHERE stat_type_id = 3;"""
        web_reg_max_obj_id = self.db_fstats.fetchquery(query)[0][0] or 0
        query = """SELECT MAX(id) FROM enum_rec_type WHERE stat_type_id = 3;"""
        web_reg_max_rec_id = self.db_fstats.fetchquery(query)[0][0] or 299

        if db_reg_max_id > web_reg_max_obj_id:
            query = """SELECT id, handle, name FROM registrar WHERE id > %d""" % web_reg_max_obj_id
            new_registrars = self.db.fetchquery(query)
            for i, registrar in enumerate(new_registrars):
                query = """INSERT INTO enum_rec_type
                            (id, rec_type_desc, stat_type_id, display_nil, collect_oper, object_id, position)
                            VALUES (%(rec_id)s, '%(reg_name)s', 3, 2, 0, %(reg_obj_id)s, 1);
                        """ % {'rec_id': web_reg_max_rec_id + i + 1,
                               'reg_name': registrar['name'],
                               'reg_obj_id': registrar['id']}
                self.db_fstats.execute(query)
                self.db_fstats.execute('COMMIT;')

        query = """SELECT MAX(id) FROM registrar;"""
        db_reg_max_id = self.db.fetchquery(query)[0][0]
        query = """SELECT MAX(object_id) FROM enum_rec_type WHERE stat_type_id = 6;"""
        web_reg_max_obj_id = self.db_fstats.fetchquery(query)[0][0] or 0
        query = """SELECT MAX(id) FROM enum_rec_type WHERE stat_type_id = 6;"""
        web_reg_max_rec_id = self.db_fstats.fetchquery(query)[0][0] or 599

        if db_reg_max_id > web_reg_max_obj_id:
            query = """SELECT id, handle, name FROM registrar WHERE id > %d""" % web_reg_max_obj_id
            new_registrars = self.db.fetchquery(query)
            for i, registrar in enumerate(new_registrars):
                query = """INSERT INTO enum_rec_type
                            (id, rec_type_desc, stat_type_id, display_nil, collect_oper, object_id, position)
                            VALUES (%(rec_id)s, '%(reg_name)s', 6, 2, 0, %(reg_obj_id)s, 1);
                        """ % {'rec_id': web_reg_max_rec_id + i + 1,
                               'reg_name': registrar['name'],
                               'reg_obj_id': registrar['id']}
                self.db_fstats.execute(query)
                self.db_fstats.execute('COMMIT;')

    def check_registrars(self):
        query = """SELECT id, name FROM registrar"""
        registrars = self.db.fetchquery(query)
        query = """UPDATE enum_rec_type SET rec_type_desc = %s
                      WHERE stat_type_id IN (3, 6)
                        AND object_id = %s
                        AND rec_type_desc <> %s"""

        for registrar in registrars:
            data = (registrar['name'], registrar['id'], registrar['name'])
            self.db_fstats.execute(query, data)
            self.db_fstats.execute('COMMIT')

    def get_number_of_domains(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s' AS value_date, 101, COUNT(*) AS value, D.zone, NOW()
                FROM
                    object_registry O
                    JOIN domain_history D ON (D.historyid = O.crhistoryid)
                WHERE type = 3
                    AND (crdate < %(midnight)s AND (erdate IS NULL OR erdate >= %(midnight)s))
                    AND name NOT LIKE '%(monitoring_domain)s'
                GROUP BY D.zone
                ORDER BY value_date, D.zone
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_number_of_domains_in_zone(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s' AS value_date, 102, COUNT(*) AS value, D.zone, NOW()
            FROM
                object_registry O
                JOIN domain_history D ON (D.historyid = O.crhistoryid)
                LEFT JOIN object_state s ON (
                    O.id = S.object_id
                    AND S.state_id=15 and S.valid_from < %(midnight)s
                    AND (S.valid_to ISNULL OR S.valid_to >= %(midnight)s))
            WHERE
                crdate < %(midnight)s
                AND (erdate isnull OR erdate >= %(midnight)s)
                AND S.id isnull and O.type=3
            GROUP BY D.zone
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_number_of_domains_with_dnssec(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s' AS value_date, 104, COUNT(*) AS value, D.zone, NOW()
                FROM
                    object_registry O
                    JOIN domain_history D ON (D.id = O.id)
                    JOIN history H ON (H.id = D.historyid)
                WHERE type = 3
                    AND (crdate <= %(midnight)s AND (erdate IS NULL OR erdate >= %(midnight)s))
                    AND name NOT LIKE '%(monitoring_domain)s'
                    AND H.valid_from < %(midnight)s AND (H.valid_to IS NULL OR H.valid_to >= %(midnight)s)
                    AND D.keyset IS NOT NULL
                GROUP BY D.zone
                ORDER BY value_date, D.zone
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_domains_in_protection(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s' AS value_date, 103, COUNT(*) AS value, D.zone, NOW()
            FROM
                object_registry O
                JOIN domain_history D ON (D.historyid = O.crhistoryid)
                JOIN object_state S ON (
                    O.id = S.object_id
                    AND S.state_id = 9 and S.valid_from < %(midnight)s
                    AND (S.valid_to IS NULL OR S.valid_to >= %(midnight)s))
            WHERE
                crdate < %(midnight)s
                AND (erdate isnull or erdate >= %(midnight)s)
                AND o.type=3
            GROUP BY D.zone
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_number_of_other_objects(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s'::date AS value_date,
                    CASE type
                        WHEN 1 THEN 201
                        WHEN 2 THEN 203
                    END, COUNT(*) AS value, 0, NOW()
                FROM
                    object_registry O
                WHERE type in (1,2)
                    AND (crdate < %(midnight)s AND (erdate IS NULL OR erdate >= %(midnight)s))
                GROUP BY type
                ORDER BY value_date
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_number_of_deleted_other_objects(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s'::date AS value_date,
                    CASE type
                        WHEN 1 THEN 204
                        WHEN 2 THEN 206
                    END,
                    COUNT(*) + CASE type
                        WHEN 1 THEN 78548 --pocet smazanych kontaktu z dsd
                        WHEN 2 THEN 0
                    END AS value, 0, NOW()
                FROM
                    object_registry O
                WHERE type in (1,2)
                    AND erdate < %(midnight)s
                GROUP BY type
                ORDER BY value_date
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_number_of_registrars_domains(self):
        pdict = dict(self.def_pdict)

        recid_regid = self.db_fstats.fetchquery("SELECT id, object_id FROM enum_rec_type WHERE stat_type_id = 3 AND object_id IS NOT NULL")
        reg_case = '\n'.join(['WHEN %d THEN %d' % (reg_id, rec_type_id) for rec_type_id, reg_id in recid_regid])
        pdict['reg_case'] = reg_case
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s'::date AS value_date,
                    --E.id,
                    CASE OBJ.clid
                        %(reg_case)s
                        --ELSE OBJ.clid
                    END as rec_type_id,
                    COUNT(*) AS value, D.zone, NOW()
                FROM
                    object_registry O
                    JOIN domain_history D ON (D.id = O.id)
                    JOIN history H ON (H.id = D.historyid)
                    JOIN object_history OBJ ON (OBJ.historyid = H.id)
                    --JOIN enum_rec_type E ON (crID = object_id)
                WHERE
                    type = 3
                    AND (crdate < %(midnight)s AND (erdate IS NULL OR erdate >= %(midnight)s))
                    AND O.name NOT LIKE '%(monitoring_domain)s'
                    AND H.valid_from < %(midnight)s AND (H.valid_to IS NULL OR H.valid_to >= %(midnight)s)
                GROUP BY D.zone, OBJ.clid -- E.id
                ORDER BY value_date, rec_type_id
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_number_of_registrars_secure_domains(self):
        pdict = dict(self.def_pdict)

        recid_regid = self.db_fstats.fetchquery("SELECT id, object_id FROM enum_rec_type WHERE stat_type_id = 6 AND object_id IS NOT NULL")
        reg_case = '\n'.join(['WHEN %d THEN %d' % (reg_id, rec_type_id) for rec_type_id, reg_id in recid_regid])
        pdict['reg_case'] = reg_case
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s'::date AS value_date,
                    CASE OBJ.clid
                        %(reg_case)s
                    END as rec_type_id,
                    COUNT(*) AS value, D.zone, NOW()
                FROM
                    object_registry O
                    JOIN domain_history D ON (D.id = O.id)
                    JOIN history H ON (H.id = D.historyid)
                    JOIN object_history OBJ ON (OBJ.historyid = H.id)
                    --JOIN enum_rec_type E ON (crID = object_id)
                WHERE
                    type = 3
                    AND (crdate < %(midnight)s AND (erdate IS NULL OR erdate >= %(midnight)s))
                    AND O.name NOT LIKE '%(monitoring_domain)s'
                    AND H.valid_from < %(midnight)s AND (H.valid_to IS NULL OR H.valid_to >= %(midnight)s)
                    AND D.keyset is not null
                GROUP BY D.zone, OBJ.clid -- E.id
                ORDER BY value_date, rec_type_id
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)


    def get_number_of_registrations_by_zone(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s'::date AS value_date, 401, COUNT(*) AS value, D.zone, NOW()
                 FROM
                     object_registry O
                     JOIN domain_history D ON (D.historyid = O.crhistoryid)
                 WHERE
                     type = 3
                     AND (crdate::date = '%(date)s'::date)
                     AND O.name NOT LIKE '%(monitoring_domain)s'
                 GROUP BY value_date, D.zone ORDER BY value_date
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_number_of_deletions(self):
        pdict = dict(self.def_pdict)
        # only delteions for unpaid maintaince fee!!!!! :
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s' AS value_date, 406, COUNT(*) AS value, D.zone, NOW()
            FROM
                object_registry O
                JOIN domain_history D ON (D.historyid = O.crhistoryid)
                JOIN object_state S ON (O.id = S.object_id)
            WHERE
                erdate BETWEEN '%(date)s'::timestamp AND %(midnight)s
                AND S.state_id = 17 and valid_to IS NULL -- deleteCandidate which was not canceled (e.g. unblocking after law suits)
                AND O.type = 3
            GROUP BY D.zone
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_domains_entered_protection(self): # aka expired
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s' AS value_date, 408, COUNT(*) AS value, D.zone, NOW()
            FROM
                object_registry O
                JOIN domain_history D ON (D.historyid = O.crhistoryid)
                JOIN object_state S ON (
                    O.id = S.object_id
                    AND S.state_id = 9 and S.valid_from::date = '%(date)s'::date
                )
            WHERE
                crdate < %(midnight)s
                AND (erdate isnull or erdate >= %(midnight)s)
                AND o.type=3
            GROUP BY D.zone
        """ % pdict
        if config.DEBUG:
            print query
        return self.db.fetchquery(query)

    def get_operations_in_registry(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s'::date AS value_date,
                    CASE r.request_type_id
                       WHEN 506 THEN 502
                       WHEN 504 THEN 503
                       WHEN 204 THEN 504
                       WHEN 404 THEN 506
                       WHEN 503 THEN 507
                       WHEN 203 THEN 508
                       WHEN 403 THEN 510
                       WHEN 505 THEN 511
                       WHEN 502 THEN 512
                    END,
                    COUNT(*), 0, NOW()
                FROM request r
                     JOIN result_code rc ON r.result_code_id = rc.id
                WHERE
                    %(log_req_cond)s
                    AND r.request_type_id in (506, 504, 204, 404, 503, 203, 403, 505, 502)
                    AND rc.result_code = 1000
                GROUP BY r.request_type_id
                ORDER BY value_date, r.request_type_id
            """ % pdict
        if config.DEBUG:
            print query
        return self.dblog.fetchquery(query)

    def get_operations_errors(self):
        pdict = dict(self.def_pdict)
        query = """
            SELECT %(year)s, %(month)s, %(day)s, '%(date)s'::date AS value_date, 513, COUNT(*), 0, NOW()
                FROM request r
                     JOIN result_code rc ON r.result_code_id = rc.id
                WHERE
                    %(log_req_cond)s
                    AND rc.result_code >= 2000
                    AND r.request_type_id IN (506, 504, 204, 404, 503, 203, 403, 505, 502)
                ORDER BY value_date
            """ % pdict
        if config.DEBUG:
            print query
        return self.dblog.fetchquery(query)

    def upload_data_to_fred_stats(self, data):
        self.db_fstats.execute(data)
        #for db_num in range(1,2): # 3
            #print 'connecting to db%d' % db_num
            #db = DBConn(db_num)
            #db.execute(data)
            #print 'data inserted to db%d' % db_num
        #print 'Data inserted successfuly.'


def main():
    if len(sys.argv) >= 2 :
        arg = sys.argv[1]
    else:
        # if no date specified, then use tomorrow date:
        d = datetime.date(1982, 1, 1).today() + datetime.timedelta(-1) # date object have today() method, so we must create random date and and then use its today()  :-(
        arg = '%d-%02d-%02d' % (d.year, d.month, d.day)
    #print arg
    if arg:
        try:
            if not re.compile('^\d{4}-\d{2}-\d{2}$').match(arg):
                raise ValueError
            comp_date = arg.split('-')
            dat = []
            for i in range(3):
                dat.append(int(comp_date[i]))
            if not (1 <= dat[1] <= 12 and 1 <= dat[2] <= 31):
                raise ValueError()

        except (IndexError, ValueError):
            sys.exit('ERROR: Argument must be in fromat YYYY-MM-DD !')
        else:
            time_total_start = time.time()

            log_file.write('%s:\n' % arg)
            if config.DEBUG:
                print '%s:' % arg
            adata = AddData()
            data = adata.get_data(arg)
            #print data
            time_measure(adata.upload_data_to_fred_stats, data)

            time_total_diff = time.time() - time_total_start
            log_file.write('%-40s %10.3f\n' % ('Total time running script:', time_total_diff))
            log_file.write('\n')
            if config.DEBUG:
                print '%-40s %10.3f' % ('Total time running script:', time_total_diff)
    else:
        sys.exit('ERROR: Program takes one optional argument in format YYYY-MM-DD !')
