CREATE OR REPLACE FUNCTION year(timestamp)
RETURNS int AS $$
SELECT EXTRACT(year FROM $1)::int;
$$ LANGUAGE sql;


-- Dopocitani total hodnot pro statistiky "Operace v registru"
DELETE FROM stat
  WHERE rec_type_id = 501
    AND year(CURRENT_DATE - 1) = year;

INSERT INTO stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) (
   SELECT year, month, day, value_date, 501, SUM(value), zone_id, MIN(inserted_date)
      FROM stat
     WHERE rec_type_id between 502 and 599 AND year = year(CURRENT_DATE - 1)
     GROUP BY  year, month, day, value_date, zone_id
);


-- ===ZONOVE SUMY===
-- Dopocitani sum nad vsemi zonamy (pouze statistiky, ktere jsou zonove zavisle)
DELETE FROM stat WHERE ((rec_type_id > 100 AND rec_type_id < 200) OR (rec_type_id > 300 AND rec_type_id < 400) OR (rec_type_id > 400 AND rec_type_id < 500) OR (rec_type_id > 600 AND rec_type_id < 800)) AND zone_id = -1 AND year = year(CURRENT_DATE - 1);
INSERT INTO stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) (
  SELECT year, month, day, value_date, rec_type_id, SUM(value), -1, NOW()
  FROM stat
  WHERE ((rec_type_id > 100 AND rec_type_id < 200) OR (rec_type_id > 300 AND rec_type_id < 400) OR (rec_type_id > 400 AND rec_type_id < 500) OR (rec_type_id > 600 AND rec_type_id < 800)) AND zone_id > 0 AND year = year(CURRENT_DATE - 1)
  GROUP BY  year, month, day, value_date, rec_type_id
);


-- ===CASOVE SUMY===

-- Dopocitani mesicnich sum
-- - collect_oper = 0 -> posledni hodnoty v mesici/roce
DELETE FROM stat WHERE day = 0 AND rec_type_id in (SELECT id FROM enum_rec_type WHERE collect_oper = 0) AND year = year(CURRENT_DATE - 1);
INSERT INTO stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) (
  SELECT year, month, 0, NULL, rec_type_id,
    (SELECT value
       FROM stat S2
       WHERE S2.day = MAX(S1.day) AND S2.year = S1.year AND S2.month = S1.month AND S2.rec_type_id = S1.rec_type_id AND S2.zone_id = S1.zone_id
    ) AS value,
    zone_id, NOW()
  FROM stat S1 JOIN enum_rec_type ON (rec_type_id = enum_rec_type.id)
  WHERE  collect_oper = 0 AND month != 0 and day != 0 AND year = year(CURRENT_DATE - 1)
  GROUP BY year, month, rec_type_id, zone_id
);


-- - collect_oper = 1 -> sumy vsech hodnoty v mesici/roce
DELETE FROM stat WHERE day = 0 AND rec_type_id in (SELECT id FROM enum_rec_type WHERE collect_oper = 1) AND year = year(CURRENT_DATE - 1);
INSERT INTO stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) (
  SELECT year, month, 0, NULL, rec_type_id, SUM(value), zone_id, NOW()
  FROM stat JOIN enum_rec_type ON (rec_type_id = enum_rec_type.id)
  WHERE  collect_oper = 1 AND month != 0 and day != 0 AND year = year(CURRENT_DATE - 1)
  GROUP BY year, month, rec_type_id, zone_id
);


-- Dopocitani rocnich sum
-- - collect_oper = 0 -> posledni hodnoty v mesici/roce
DELETE FROM stat WHERE month = 0 AND day = 0 AND rec_type_id in (SELECT id FROM enum_rec_type WHERE collect_oper = 0) AND year = year(CURRENT_DATE - 1);
INSERT INTO stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) (
  SELECT year, 0, 0, NULL, rec_type_id,
    (SELECT value
       FROM stat S2
       WHERE S2.month = MAX(S1.month) AND S2.year = S1.year AND S2.day = 0 AND S2.rec_type_id = S1.rec_type_id AND S2.zone_id = S1.zone_id
    ),
    zone_id, NOW()
  FROM stat S1 JOIN enum_rec_type ON (rec_type_id = enum_rec_type.id)
  WHERE  collect_oper = 0 AND month != 0 and day = 0 AND year = year(CURRENT_DATE - 1)
  GROUP BY year, day, rec_type_id, zone_id
);

-- - collect_oper = 1 -> sumy vsech hodnoty v mesici/roce
DELETE FROM stat WHERE month = 0 AND day = 0 AND rec_type_id in (SELECT id FROM enum_rec_type WHERE collect_oper = 1) AND year = year(CURRENT_DATE - 1);
INSERT INTO stat (year, month, day, value_date, rec_type_id, value, zone_id, inserted_date) (
  SELECT year, 0, 0, NULL, rec_type_id, SUM(value), zone_id, NOW()
  FROM stat JOIN enum_rec_type ON (rec_type_id = enum_rec_type.id)
  WHERE  collect_oper = 1 AND month != 0 and day = 0 AND year = year(CURRENT_DATE - 1)
  GROUP BY year, rec_type_id, zone_id
);
